
const flatten = require('lodash.flatten')
const anymatch = require('anymatch')
const Lev = require('./level-wrapper')
const {debug, veryDebug, lock} = require('./util')

// exported api
// =

exports.loadArchives = async function (db, needsRebuild) {
  debug('Indexer.loadArchives, needsRebuild=' + needsRebuild)
  var promises = []
  await db._indexMeta.each(indexMeta => {
    debug('loading archive', indexMeta._url, indexMeta.localPath)
    // load the archive
    const archive = new (db.DatArchive)(indexMeta._url, {localPath: indexMeta.localPath})
    archive.isWritable = indexMeta.isWritable
    if (archive.isWritable) db.prepareArchive(archive)
    db._archives[archive.url] = archive

    // process the archive
    promises.push(indexArchive(db, archive, needsRebuild))
    exports.watchArchive(db, archive)
  })
  await Promise.all(promises)
  debug('Indexer.loadArchives done')
}

exports.addArchive = async function (db, archive) {
  veryDebug('Indexer.addArchive', archive.url)
  // store entry in the meta db
  var info = await archive.getInfo()
  archive.isWritable = info.isOwner
  await db._indexMeta.level.put(archive.url, {
    _url: archive.url,
    version: 0,
    isWritable: archive.isWritable,
    localPath: archive._localPath
  })
  // process the archive
  await indexArchive(db, archive)
  exports.watchArchive(db, archive)
}

exports.removeArchive = async function (db, archive) {
  veryDebug('Indexer.removeArchive', archive.url)
  await unindexArchive(db, archive)
  exports.unwatchArchive(db, archive)
}

exports.watchArchive = async function (db, archive) {
  veryDebug('Indexer.watchArchive', archive.url)
  if (archive.fileEvents) {
    console.error('watchArchive() called on archive that already is being watched', archive.url)
    return
  }
  if (archive._loadPromise) {
    // HACK node-dat-archive fix
    // Because of a weird API difference btwn node-dat-archive and beaker's DatArchive...
    // ...the event-stream methods need await _loadPromise
    // -prf
    await archive._loadPromise
  }
  archive.fileEvents = archive.createFileActivityStream(db.tablePathPatterns)
  // autodownload all changes to the watched files
  archive.fileEvents.addEventListener('invalidated', ({path}) => archive.download(path))
  // autoindex on changes
  // TODO debounce!!!!
  archive.fileEvents.addEventListener('changed', ({path}) => indexArchive(db, archive))
}

exports.unwatchArchive = function (db, archive) {
  veryDebug('unwatching', archive.url)
  if (archive.fileEvents) {
    archive.fileEvents.close()
    archive.fileEvents = null
  }
}

exports.waitTillIndexed = async function (db, archive) {
  debug('Indexer.waitTillIndexed', archive.url)
  // fetch the current state of the archive's index
  var [indexMeta, archiveMeta] = await Promise.all([
    db._indexMeta.level.get(archive.url),
    archive.getInfo()
  ])
  indexMeta = indexMeta || {version: 0}

  // done?
  if (indexMeta.version >= archiveMeta.version) {
    debug('Indexer.waitTillIndexed already indexed')
    return
  }

  return new Promise(resolve => {
    db.on('indexes-updated', onIndex)
    function onIndex (indexedArchive, version) {
      if (indexedArchive.url === archive.url && version >= archiveMeta.version) {
        db.removeListener('indexes-updated', onIndex)
        resolve()
      }
    }
  })
}

exports.resetOutdatedIndexes = async function (db) {
  if (db._tablesToRebuild.length === 0) {
    return false
  }
  debug(`Indexer.resetOutdatedIndexes need to rebuid ${db._tablesToRebuild.length} tables`)
  veryDebug('Indexer.resetOutdatedIndexes tablesToRebuild', db._tablesToRebuild)

  // clear tables
  // TODO
  // for simplicity, we just clear all data and re-index everything
  // a better future design would only clear the tables that changed
  // unfortunately our indexer isn't smart enough for that yet
  // -prf
  const tables = db.tables
  for (let i = 0; i < tables.length; i++) {
    let table = tables[i]
    veryDebug('clearing', table.name)
    // clear indexed data
    await Lev.clear(table.level)
  }

  // reset meta records
  var promises = []
  await db._indexMeta.each(indexMeta => {
    indexMeta.version = 0
    promises.push(db._indexMeta.level.put(indexMeta._url, indexMeta))
  })
  await Promise.all(promises)

  return true
}

// figure how what changes need to be processed
// then update the indexes
async function indexArchive (db, archive, needsRebuild) {
  debug('Indexer.indexArchive', archive.url, {needsRebuild})
  var release = await lock(`index:${archive.url}`)
  try {
    // sanity check
    if (!db.isOpen) {
      return
    }
    if (!db.level) {
      return console.log('indexArchive called on corrupted db')
    }

    // fetch the current state of the archive's index
    var [indexMeta, archiveMeta] = await Promise.all([
      db._indexMeta.level.get(archive.url).catch(e => null),
      archive.getInfo()
    ])
    indexMeta = indexMeta || {version: 0}

    // has this version of the archive been processed?
    if (indexMeta && indexMeta.version >= archiveMeta.version) {
      debug('Indexer.indexArchive no index needed for', archive.url)
      return // yes, stop
    }
    debug('Indexer.indexArchive ', archive.url, 'start', indexMeta.version, 'end', archiveMeta.version)

    // find and apply all changes which haven't yet been processed
    var updates = await scanArchiveHistoryForUpdates(db, archive, {
      start: indexMeta.version + 1,
      end: archiveMeta.version + 1
    })
    var results = await applyUpdates(db, archive, archiveMeta, updates)
    debug('Indexer.indexArchive applied', results.length, 'updates from', archive.url)

    // update meta
    await Lev.update(db._indexMeta.level, archive.url, {
      _url: archive.url,
      version: archiveMeta.version // record the version we've indexed
    })

    // emit
    var updatedTables = new Set(results)
    for (let tableName of updatedTables) {
      if (!tableName) continue
      db[tableName].emit('index-updated', archive, archiveMeta.version)
    }
    db.emit('indexes-updated', archive, archiveMeta.version)
  } finally {
    release()
  }
}
exports.indexArchive = indexArchive

// delete all records generated from the archive
async function unindexArchive (db, archive) {
  var release = await lock(`index:${archive.url}`)
  try {
    // find any relevant records and delete them from the indexes
    var recordMatches = await scanArchiveForRecords(db, archive)
    await Promise.all(recordMatches.map(match => match.table.level.del(match.recordUrl)))
    await db._indexMeta.level.del(archive.url)
  } finally {
    release()
  }
}
exports.unindexArchive = unindexArchive

// internal methods
// =

// look through the given history slice
// match against the tables' path patterns
// return back the *latest* change to each matching changed record
async function scanArchiveHistoryForUpdates (db, archive, {start, end}) {
  var history = await archive.history({start, end})
  var updates = {}
  history.forEach(update => {
    if (anymatch(db._tablePathPatterns, update.path)) {
      updates[update.path] = update
    }
  })
  return updates
}

// look through the archive for any files that generate records
async function scanArchiveForRecords (db, archive) {
  var recordFiles = await Promise.all(db.tables.map(table => {
    return table.listRecordFiles(archive)
  }))
  return flatten(recordFiles)
}

// iterate the updates and apply them to the indexes
async function applyUpdates (db, archive, archiveMeta, updates) {
  return Promise.all(Object.keys(updates).map(async path => {
    var update = updates[path]
    if (update.type === 'del') {
      return unindexFile(db, archive, update.path)
    } else {
      return readAndIndexFile(db, archive, archiveMeta, update.path)
    }
  }))
}

// read the file, find the matching table, validate, then store
async function readAndIndexFile (db, archive, archiveMeta, filepath) {
  const tables = db.tables
  const fileUrl = archive.url + filepath
  try {
    // read file
    var record = JSON.parse(await archive.readFile(filepath))

    // index on the first matching table
    for (var i = 0; i < tables.length; i++) {
      let table = tables[i]
      if (table.isRecordFile(filepath)) {
        // validate if needed
        if (table.schema.validator) {
          record = table.schema.validator(record)
        }
        // add standard attributes
        record._url = fileUrl
        record._origin = archive.url
        record._author = archiveMeta && archiveMeta.author && archiveMeta.author.url
        // save
        await table.level.put(record._url, record)
        return table.name
      }
    }
  } catch (e) {
    console.log('Failed to index', fileUrl, e)
  }
  return false
}

async function unindexFile (db, archive, filepath) {
  const tables = db.tables
  const fileUrl = archive.url + filepath
  try {
    // unindex on the first matching table
    for (var i = 0; i < tables.length; i++) {
      let table = tables[i]
      if (table.isRecordFile(filepath)) {
        await table.level.del(fileUrl)
        return table.name
      }
    }
  } catch (e) {
    console.log('Failed to unindex', fileUrl, e)
  }
  return false
}
