const Lev = require('./level-wrapper')
const InjestWhereClause = require('./where-clause')
const Indexer = require('./indexer')
const {assert, debug} = require('./util')
const {QueryError, ParameterError} = require('./errors')

class InjestQuery {
  constructor (table) {
    this._table = table
    this._filters = []
    this._reverse = false
    this._offset = 0
    this._limit = false
    this._until = null
    this._where = null
  }

  // () => InjestQuery
  clone () {
    var clone = new InjestQuery()
    for (var k in this) {
      if (k.startsWith('_')) {
        clone[k] = this[k]
      }
    }
    return clone
  }

  // () => Promise<Number>
  async count () {
    var count = 0
    await this.each(() => { count++ })
    return count
  }

  // () => Promise<Number>
  async delete () {
    var deletes = []
    await this.each(record => {
      const archive = this._table.db._archives[record._origin]
      debug('InjestQuery.delete', record)
      if (archive && archive.isWritable) {
        const filepath = record._url.slice(record._origin.length)
        deletes.push(
          archive.unlink(filepath)
            .then(() => Indexer.indexArchive(this._table.db, archive))
        )
      } else {
        debug('InjestQuery.delete not enacted:', !archive ? 'Archive not found' : 'Archive not writable')
      }
    })
    await Promise.all(deletes)
    return deletes.length
  }

  // (Function) => Promise<Void>
  async each (fn) {
    return Lev.iterate(this, fn)
  }

  // (Function) => Promise<Void>
  async eachKey (fn) {
    assert(typeof fn === 'function', ParameterError, `First parameter of .eachKey() must be a function, got ${fn}`)
    return this.each(cursor => {
      // choose the key
      var key = this._table.schema.primaryKey // default to the primary key
      if (this._where && this._where._index) {
        key = this._where._index // use the where clause's key if there is one
      }
      if (!key) key = '_url' // fallback to url

      // emit all
      if (Array.isArray(cursor[key])) {
        cursor[key].forEach(v => fn(v))
      } else {
        fn(cursor[key])
      }
    })
  }

  // (Function) => Promise<Void>
  async eachUrl (fn) {
    assert(typeof fn === 'function', ParameterError, `First parameter of .eachUrl() must be a function, got ${fn}`)
    return this.each(cursor => { fn(cursor._url) })
  }

  // (Function) => InjestQuery
  filter (fn) {
    assert(typeof fn === 'function', ParameterError, `First parameter of .filter() must be a function, got ${fn}`)
    this._filters.push(fn)
    return this
  }

  // () => Promise<Object>
  async first () {
    var arr = await this.limit(1).toArray()
    return arr[0]
  }

  // () => Promise<Array<String>>
  async keys () {
    var keys = []
    await this.eachKey(key => keys.push(key))
    return keys
  }

  // () => Promise<Object>
  async last () {
    return this.reverse().first()
  }

  // (Number) => InjestQuery
  limit (n) {
    assert(typeof n === 'number', ParameterError, `The first parameter to .limit() must be a number, got ${n}`)
    this._limit = n
    return this
  }

  // (Number) => InjestQuery
  offset (n) {
    assert(typeof n === 'number', ParameterError, `The first parameter to .offset() must be a number, got ${n}`)
    this._offset = n
    return this
  }

  // (index) => InjestWhereClause
  or (index) {
    assert(this._where, QueryError, 'Can not have a .or() before a .where()')
    // TODO
  }

  // (index) => InjestQuery
  orderBy (index) {
    assert(typeof index === 'string', ParameterError, `The first parameter to .orderBy() must be a string, got ${index}`)
    assert(!this._where, QueryError, 'Can not have an .orderBy() and a .where() - where() implicitly sets the orderBy() to its key')
    this._where = new InjestWhereClause(this, index)
    return this
  }

  // () => Promise<Array<String>>
  async urls () {
    var urls = []
    await this.eachUrl(url => urls.push(url))
    return urls
  }

  // () => InjestQuery
  reverse () {
    this._reverse = true
    return this
  }

  // () => Promise<Array<Object>>
  async toArray () {
    var records = []
    await this.each(record => records.push(record))
    return records
  }

  // () => Promise<Array<String>>
  async uniqueKeys () {
    return Array.from(new Set(await this.keys()))
  }

  // (Function) => InjestQuery
  until (fn) {
    assert(typeof fn === 'function', ParameterError, `First parameter of .until() must be a function, got ${fn}`)
    this._until = fn
    return this
  }

  // (Object|Function) => Promise<Number>
  async update (objOrFn) {
    var fn
    if (objOrFn && typeof objOrFn === 'object') {
      // create a function which applies the object updates
      const obj = objOrFn
      fn = record => {
        for (var k in obj) {
          if (k === '_url' || k === '_origin' || k === '_author') {
            continue // skip special attrs
          }
          if (typeof obj[k] !== 'undefined') {
            record[k] = obj[k]
          }
        }
      }
    } else if (typeof objOrFn === 'function') {
      fn = objOrFn
    } else {
      throw new ParameterError(`First parameter of .update() must be a function or object, got ${objOrFn}`)
    }

    // apply updates
    var updates = []
    await this.each(record => {
      const archive = this._table.db._archives[record._origin]
      debug('InjestQuery.update', record)
      if (archive && archive.isWritable) {
        // run update
        fn(record)

        // run validation
        if (this._table.schema.validator) {
          let {_url, _origin, _author} = record
          record = this._table.schema.validator(record)
          record._url = _url
          record._origin = _origin
          record._author = _author
        }

        // run to-file pass
        const fileData = (this._table.schema.toFile) ?
          this._table.schema.toFile(record) :
          record

        // write to archive
        const filepath = record._url.slice(record._origin.length)
        updates.push(
          archive.writeFile(filepath, JSON.stringify(fileData))
            .then(() => {
              if (typeof archive.commit === 'function') {
                // legacy dat api
                return archive.commit()
              }
            })
            .then(() => Indexer.indexArchive(this._table.db, archive))
        )
      } else {
        debug('InjestQuery.delete not enacted:', !archive ? 'Archive not found' : 'Archive not writable')
      }
    })
    await Promise.all(updates)
    return updates.length
  }

  // (index|query) => InjestWhereClause|InjestQuery
  where (indexOrQuery) {
    assert(!this._where, QueryError, 'Can not have two .where()s unless they are separated by a .or()')
    this._where = new InjestWhereClause(this, indexOrQuery)
    return this._where
  }
}

module.exports = InjestQuery
